# this forge file is split out from root.py so that projects built with forge
# can use this configuration information
import os
import sys

forge = sys.modules["forge"]

forge.cprint(forge.YELLOW,0,'Free Electron Common forge.py support')

def platform_setup():
	if forge.api == "x86_linux" or forge.api == "x86_64_linux":
		forge.fe_os = 'FE_LINUX'
		forge.fe_osver = '2'
		forge.fe_compiler = 'FE_GNU'
		forge.fe_gl2d = 'FE_2D_X_GFX'
		forge.fe_gl3d = 'FE_3D_OPENGL'
		forge.fe_hw = 'FE_X86'
		forge.includemap['include'] = '/usr/X11R6/include'
		forge.linkmap['linux'] = '-lm -lc -lcrypt -ldl'
		forge.winlibs = '-L/usr/X11/lib -L/usr/X11R6/lib -lX11 -lXext'
		if forge.api == "x86_64_linux":
			forge.winlibs = '-L/usr/X11R6/lib64 ' + forge.winlibs
		forge.gllibs = '-lGL'
		forge.gllibs += ' -lGLU'
		forge.cppmap['checks'] = '-ansi'
		forge.linkmap['checks'] = '-ansi'
		forge.cppmap['efficiency'] = '-pipe'
		forge.linkmap['efficiency'] = '-pipe'
		forge.cppmap['multithread'] = '-D_REENTRANT -D_POSIX_REENTRANT_FUNCTIONS -DZT_POSIX'
		forge.linkmap['multithread'] = '-lpthread -lrt'
		forge.linkmap['rpath'] = " -Xlinker -rpath='$ORIGIN/'"
		forge.cppmap['template-depth'] = '-ftemplate-depth-60'
	elif forge.api == "ppc_linux":
		forge.fe_os = 'FE_LINUX'
		forge.fe_osver = '2'
		forge.fe_compiler = 'FE_GNU'
		forge.fe_gl2d = 'FE_2D_X_GFX'
		forge.fe_gl3d = 'FE_3D_OPENGL'
		forge.fe_hw = 'FE_PPC'
		forge.includemap['include'] = '/usr/X11R6/include'
		forge.linkmap['linux'] = '-lm -lc -lcrypt -ldl'
		forge.winlibs = '-L/usr/X11/lib -L/usr/X11R6/lib -lX11 -lXext'
		forge.gllibs = '-lGL'
		forge.gllibs += ' -lGLU'
		forge.cppmap['checks'] = '-ansi'
		forge.linkmap['checks'] = '-ansi'
		forge.cppmap['efficiency'] = '-pipe'
		forge.linkmap['efficiency'] = '-pipe'
		forge.cppmap['multithread'] = '-D_REENTRANT -D_POSIX_REENTRANT_FUNCTIONS -DZT_POSIX'
		forge.linkmap['multithread'] = '-lpthread -lrt'
		forge.cppmap['template-depth'] = '-ftemplate-depth-60'
	elif forge.api == "ppc_osx" or forge.api == "x86_osx":
		forge.fe_os = 'FE_OSX'
		forge.fe_osver = '10'
		forge.fe_compiler = 'FE_GNU'
#		forge.fe_gl2d = 'FE_2D_OSX'
		forge.fe_gl2d = 'FE_2D_X_GFX'
		forge.fe_gl3d = 'FE_3D_OPENGL'
		forge.fe_hw = 'FE_PPC'
		forge.includemap['include'] = '/usr/X11R6/include'
		forge.linkmap['osx'] = '-lm -lc -ldl'
		forge.winlibs = '-L/usr/X11R6/lib -lX11 -lXext'
#		forge.gllibs = '-framework OpenGL -framework GLUT'
		forge.gllibs += ' -lGL -lGLU'
		forge.cppmap['checks'] = '-ansi'
		forge.linkmap['checks'] = '-ansi'
		forge.cppmap['efficiency'] = '-pipe'
		forge.linkmap['efficiency'] = '-pipe'
		forge.cppmap['multithread'] = '-D_REENTRANT -D_POSIX_REENTRANT_FUNCTIONS -DZT_POSIX'
		forge.linkmap['multithread'] = '-lpthread'
		forge.cppmap['template-depth'] = '-ftemplate-depth-60'

	elif forge.api == "x86_cygwin":
		forge.fe_os = 'FE_CYGWIN'
		forge.fe_osver = '2'
		forge.fe_compiler = 'FE_GNU'
		forge.fe_gl2d = 'FE_2D_X_GFX'
		forge.fe_gl3d = 'FE_3D_OPENGL'
		forge.fe_hw = 'FE_X86'
		forge.includemap['include'] = '/usr/X11R6/include'
		forge.linkmap['linux'] = '-lm -lc -lcrypt -ldl'
		forge.winlibs = '-L/usr/X11/lib -L/usr/X11R6/lib -lX11 -lXext'
		forge.gllibs = '-lGL -lGLU'
		forge.cppmap['multithread'] = ' -DREENTRANT -DZT_POSIX'
		forge.linkmap['multithread'] += '-lpthread -lrt'

	elif forge.api == "x86_win32" or forge.api == "x86_win64":
		if forge.api == "x86_win64":
			forge.fe_os = 'FE_WIN64'
		else:
			forge.fe_os = 'FE_WIN32'
		forge.fe_osver = '4'
		forge.fe_compiler = 'FE_MICROSOFT'
		forge.fe_gl2d = 'FE_2D_GDI'
		forge.fe_gl3d = 'FE_3D_OPENGL'
		forge.fe_hw = 'FE_X86'
		forge.linkmap['win32'] = 'kernel32.lib user32.lib gdi32.lib winspool.lib comdlg32.lib advapi32.lib shell32.lib ole32.lib oleaut32.lib uuid.lib winmm.lib ws2_32.lib'
		forge.winlibs = ''
		forge.gllibs = 'opengl32.lib'
		forge.gllibs += ' glu32.lib'
		forge.cprint(forge.RED,1,'downgrading warning level from /W3 to /W1 due to boost slop')
		forge.cppmap['win32_warning'] = '/W1'

	forge.apiCodegen = forge.api + "_" + forge.codegen

def compiler_setup():
	compiler = os.getenv('FE_CC')
	if compiler:
		compiler_base = os.path.basename(compiler)
#		forge.cxx = compiler		overwrite setting from forge.py
		if compiler_base == 'icc':
			forge.fe_compiler = 'FE_INTEL'
		if compiler_base == 'dmc':
			forge.cxx = 'C:\\dm\\bin\\sc.exe'
			forge.fe_compiler = 'FE_DMC'

	if forge.fe_compiler == 'FE_INTEL':
		forge.cppmap['codegen'] = ''
		if forge.codegen == 'debug':
			forge.cppmap['intel'] = '-g'
		forge.linkmap['intel_cc'] = '-lcxa -lcprts -lunwind'

	if forge.fe_compiler == 'FE_MICROSOFT':
		if forge.codegen == 'debug':
			forge.cppmap['fe_codegen'] = '-DFE_CONSOLE_LOG'
		forge.linkmap['win32_no_cmt'] = '/NODEFAULTLIB:libcmt.lib /NODEFAULTLIB:libcmtd.lib'

	if forge.fe_compiler == 'FE_DMC':
		forge.cprint(forge.RED,1,'Digital Mars build incomplete. Awaiting full boost support')
		sys.exit(1)
		forge.objout = '-o'
		forge.exeout = '-o'
		forge.dlflags = ' -WD'
		if forge.codegen == 'debug':
			forge.cppmap['codegen'] = '-D_DEBUG -g'
			forge.dlflags += ' /DEBUG'
		elif forge.codegen == 'optimize':
			forge.cppmap['codegen'] = '-DNDEBUG'
		else:
			forge.cppmap['codegen'] = ''
		#forge.cppmap['codegen'] += ' -Aa -Ab -Ae -Ar -Aw -C -EC -Jb -Pz -3 -mn'
		forge.cppmap['codegen'] += ' -A- -Aa -Ae -Ar'
		forge.libincl = '-L'
		forge.nolink = '-c'
		forge.linker = 'c:\\dm\\bin\dmc.exe'
		forge.linkerflags = ''
		forge.ar = 'c:\\dm\\bin\\lib.exe'
		forge.arflags = '-p32'
		forge.aroutprefix = '-c '
		forge.objsuf = '.obj'
		forge.exesuf = '.exe'
		forge.libflag = ''
		forge.libpre = ''
		forge.libsuf = '.lib'
		forge.libdl = '.dll'
		forge.pymodsuf = '.pyd'
		forge.ld1 = ''
		forge.ld2 = ''
		forge.forcecxx = '-cpp'
		forge.fat1 = ''
		forge.fat2 = ''

		forge.winlibs = ''
		forge.gllibs = ''

		forge.cppmap['win32'] =''
		forge.cppmap['win32_warning'] = ''

		forge.includemap['stl'] = 'C:\dm\stlport\stlport'
		forge.linkmap['win32'] = 'kernel32.lib user32.lib gdi32.lib comdlg32.lib advapi32.lib ole32.lib oleaut32.lib uuid.lib'
		#forge.includemap['dmc'] = 'C:\dm\include'

	forge.gfxlibs = forge.winlibs + ' ' + forge.gllibs

def scan_houdini_installs():
	forge.houdini_sources = []
	houdini_boost = []
	houdini_dir = forge.houdini_installs
	houdini_path = os.getenv('HT')
	if houdini_path:
		houdini_dir = os.path.normpath(houdini_path + '/../..') + '/'

	houdini_found = {}

	for version in forge.houdini_versions:
		for prefix in [ '', 'hfs' ]:
			for suffix in [ '', '_64' ]:
				houdini_version = version[0] + suffix
				boost_version = version[1]
				version_path = houdini_dir + prefix + houdini_version
				if os.path.exists(version_path + '/dsolib'):
					vpti = version_path + '/toolkit/include'
					vpd = version_path + '/dsolib'
					forge.houdini_sources += [[ houdini_version, vpti, vpd ]]
					if not boost_version in houdini_found:
						houdini_boost += [[ boost_version, vpti, vpd ]]
						houdini_found[ boost_version ] = houdini_version

	forge.boost_sources += houdini_boost

	if len(forge.houdini_sources) == 0 and houdini_path:
		forge.houdini_sources = [[ '',
				os.path.join(houdini_path, 'include'), os.getenv('HDSO') ]]

def scan_maya_installs():
	forge.maya_sources = []
	maya_dir = forge.maya_installs
	maya_path = os.getenv('MAYA_ROOT')
	if maya_path:
		maya_dir = os.path.normpath(maya_path + '/..') + '/'

	for version in forge.maya_versions:
		for prefix in [ '', 'maya' ]:
			for suffix in [ '-x64', '_64', '_test_64', '' ]:
				maya_version = version + suffix
				version_path = maya_dir + prefix + maya_version
				if os.path.exists(version_path + '/include'):
					vpti = version_path + '/include'
					vpd = version_path + '/lib'
					forge.maya_sources += [[ maya_version, vpti, vpd ]]
					break

	if len(forge.maya_sources) == 0 and maya_path:
		forge.maya_sources = [[ '',
				os.path.join(maya_path, 'include'),
				os.path.join(maya_path, 'lib') ]]

def find_boost_lib(path, collection):

	for suffix in [ '-mt', '' ]:
		if os.path.exists(path + '/libboost_' + collection + suffix + '.so'):
			return 'boost_' + collection + suffix

	return ''
