/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_FCO_DEBUG		FALSE
#define FE_FCO_EDIT_DEBUG	FALSE
#define FE_FCO_OUTPUT_DEBUG	FALSE

#include "operator/operator.pmh"

using namespace fe;

void FollicleOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	//* page Data
	catalog<bool>("useUV")=false;
	catalog<String>("useUV","label")="Use UV";
	catalog<String>("useUV","page")="Data";
	catalog<String>("useUV","hint")=
			"Make changes using UV coordinates instead of"
			"simply snapping to the surface in world space.";

	catalog<bool>("affectAttribute")=false;
	catalog<String>("affectAttribute","label")="Affect Attribute";
	catalog<bool>("affectAttribute","joined")=true;
	catalog<String>("affectAttribute","page")="Data";

	catalog<String>("attribute");
	catalog<String>("attribute","label")="Attribute";
	catalog<String>("attribute","enabler")="affectAttribute";
	catalog<String>("attribute","page")="Data";

	catalog<String>("hideGroup")="hide";
	catalog<String>("hideGroup","label")="Hide Group";
	catalog<String>("hideGroup","page")="Data";
	catalog<String>("hideGroup","hint")=
			"Put all hidden points in this new group.";

	//* page Edit
	catalog<String>("modifyMode")="Drag";
	catalog<String>("modifyMode","label")="Mode";
	catalog<String>("modifyMode","choice:0")="None";
	catalog<String>("modifyMode","choice:1")="Drag";
	catalog<String>("modifyMode","choice:2")="Paint";
	catalog<bool>("modifyMode","joined")=true;
	catalog<String>("modifyMode","page")="Edit";

	catalog<String>("action")="Apply";
	catalog<String>("action","label")="Action";
	catalog<String>("action","choice:0")="Apply";
	catalog<String>("action","choice:1")="Smooth";
	catalog<String>("action","choice:2")="Restore";
	catalog<String>("action","page")="Edit";

	catalog<Real>("intensity")=0.1;
	catalog<Real>("intensity","high")=1.0;
	catalog<String>("intensity","label")="Intensity";
	catalog<String>("intensity","page")="Edit";

	catalog<Real>("pickRadius")=0.1;
	catalog<Real>("pickRadius","high")=1.0;
	catalog<String>("pickRadius","label")="Pick Radius";
	catalog<bool>("pickRadius","joined")=true;
	catalog<String>("pickRadius","page")="Edit";
	catalog<String>("pickRadius","hint")="Inner region of area select.";

	catalog<Real>("falloff")=0.2;
	catalog<String>("falloff","label")="Falloff";
	catalog<String>("falloff","page")="Edit";
	catalog<String>("falloff","hint")="Falloff extension of area select.";

	catalog<Real>("fuzzy")=0.0;
	catalog<Real>("fuzzy","max")=1.0;
	catalog<String>("fuzzy","label")="Fuzzy Weights";
	catalog<String>("fuzzy","page")="Edit";
	catalog<String>("fuzzy","hint")="Selection randomness.";

	//* page Display
	catalog<Real>("singleRadius")=0.1;
	catalog<String>("singleRadius","label")="Single Radius";
	catalog<String>("singleRadius","page")="Display";
	catalog<String>("singleRadius","hint")=
			"Drawn radius during single select.";

	catalog<I32>("lineWidth")=2;
	catalog<I32>("lineWidth","min")=1;
	catalog<I32>("lineWidth","max")=10;
	catalog<String>("lineWidth","label")="Line Width";
	catalog<String>("lineWidth","page")="Display";
	catalog<String>("lineWidth","hint")="Thickness of box lines.";

	catalog<I32>("rimSize")=4;
	catalog<I32>("rimSize","min")=2;
	catalog<I32>("rimSize","max")=10;
	catalog<String>("rimSize","label")="Rim Size";
	catalog<String>("rimSize","page")="Display";
	catalog<String>("rimSize","hint")="Width of boxes.";

	catalog< sp<Component> >("Brush");
	catalog<String>("Brush","implementation")="DrawI";
	catalog<String>("Brush","prompt")=
			"LMB selects and moves points."
			"  MMB creates points."
			"  Enter toggles area select."
			"  DEL reverts changes."
			"  shift-DEL hides."
			"  ESC deselects.";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Driver Surface");

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
	catalog<String>("Output Surface","copy")="Input Surface";
	//* TODO reuse would need to be able to delete points from mesh
//	catalog<bool>("Output Surface","recycle")=true;

	m_spWireframe=new DrawMode();
	m_spWireframe->setDrawStyle(DrawMode::e_wireframe);
	m_spWireframe->setPointSize(3.0);
	m_spWireframe->setLineWidth(2.0);

	m_spSolid=new DrawMode();
	m_spSolid->copy(m_spWireframe);
	m_spSolid->setDrawStyle(DrawMode::e_solid);
	m_spSolid->setLit(FALSE);
	m_spSolid->setBackfaceCulling(FALSE);
}

void FollicleOp::handle(Record& a_rSignal)
{
	const String modifyMode=catalog<String>("modifyMode");
	const String action=catalog<String>("action");
	const Real intensity=catalog<Real>("intensity");
	const Real singleRadius=catalog<Real>("singleRadius");
	const Real pickRadius=catalog<Real>("pickRadius");
	const Real falloff=catalog<Real>("falloff");
	const Real fuzzy=catalog<Real>("fuzzy");

	const I32 lineWidth=catalog<I32>("lineWidth");
	const I32 rimSize=catalog<I32>("rimSize");

	const BWORD affectAttribute=catalog<bool>("affectAttribute");
	const String attribute=catalog<String>("attribute");

	sp<DrawI> spDrawBrush;
	accessBrush(spDrawBrush,a_rSignal,e_quiet);

	sp<DrawI> spDrawOverlay;
	accessBrushOverlay(spDrawOverlay,a_rSignal,e_quiet);

#if FE_FCO_DEBUG
	feLog("FollicleOp::handle brush %d brushed %d\n",
			spDrawBrush.isValid(),m_brushed);
#endif

	if(spDrawBrush.isValid())
	{
		m_spWireframe->setLineWidth(lineWidth);

		spDrawBrush->pushDrawMode(m_spWireframe);
		spDrawOverlay->pushDrawMode(m_spSolid);

		m_brushed=TRUE;

		const Color black(0.0,0.0,0.0);
		const Color white(1.0,1.0,0.8);
		const Color grey(0.8,0.8,1.0);
		const Color red(0.8,0.0,0.0);
		const Color green(0.0,0.5,0.0);
		const Color blue(0.0,0.0,1.0);
		const Color cyan(0.0,1.0,1.0);
		const Color magenta(0.8,0.0,0.8);
		const Color yellow(1.0,1.0,0.0);
		const Color sparkle(1.0,1.0,0.6);
		const Color orange(1.0,0.5,0.0);
		const Color blank(0.0,0.0,0.0,0.0);

		const SpatialVector up(0.0,1.0,0.0);

		/***
			select/move			Left, Drag
			unselect			click off object
			add					middle
			hide selection		Del
			restore				ctrl-Del

			multi-select		ctrl-single, shift-area + drag
			soft-select			Falloff (dot aligned)
			area-select			Radius (dot aligned)
			fuzzy-select		Noise-Weighting

			Enter toggles single/area mode
			soft available in area mode
			fuzzy is parameter of soft
			weight color like maya soft select (yellow->red?)

			inner box - selected (color by weight)
			outline box - to select
		*/

		if(m_spDriver.isValid())
		{
			const SpatialVector& rRayOrigin=rayOrigin(a_rSignal);
			const SpatialVector& rRayDirection=rayDirection(a_rSignal);

			m_event.bind(windowEvent(a_rSignal));

#if FE_FCO_DEBUG
			feLog("%s\n",c_print(m_event));
#endif

			SpatialVector intersection;
			set(intersection);
			SpatialVector hitNormal;
			set(hitNormal);
			BWORD impacted=FALSE;

			const Real maxDistance=100.0;
			sp<SurfaceI::ImpactI> spImpact=m_spDriver->rayImpact(
					rRayOrigin,rRayDirection,maxDistance);
			if(spImpact.isValid())
			{
				sp<SurfaceTriangles::Impact> spTriImpact=spImpact;
				if(spTriImpact.isValid())
				{
					intersection=spTriImpact->intersection();
					hitNormal=spTriImpact->normal();
#if FE_FCO_DEBUG
					const I32 face=spTriImpact->face();
					const I32 pickIndex=spTriImpact->triangleIndex();
					feLog("FollicleOp::handle face %d pick %d %s\n",
							face,pickIndex,c_print(intersection));
#endif

					impacted=TRUE;
				}
			}

			sp<ViewI> spView=spDrawOverlay->view();
			sp<ViewMode> spViewMode=spView->viewMode();
			SpatialTransform camera=spViewMode->camera();
			SpatialTransform invCamera;
			invert(invCamera,camera);

			const SpatialVector deltaX=invCamera.column(0);
			const SpatialVector deltaY=invCamera.column(1);

			I32 closest= -1;
			Real closestDist= -1.0;

			const U32 outputCount=m_spOutputPoint->count();
			m_colorRim.resize(outputCount);
			m_colorCenter.resize(outputCount);

			const Real extent=pickRadius+falloff;
			const Real extent2=extent*extent;
			for(I32 pointIndex=0;pointIndex<outputCount;pointIndex++)
			{
				m_colorCenter[pointIndex]=blank;

				const SpatialVector normal=
						m_spOutputNormal->spatialVector(pointIndex);
				const Real hitDot=dot(normal,invCamera.column(2));
				if(hitDot<0)
				{
					m_colorRim[pointIndex]=blank;
					continue;
				}

				const SpatialVector point=
						m_spOutputPoint->spatialVector(pointIndex);

				m_colorRim[pointIndex]=green;

				if(spImpact.isValid())
				{
					const Real dist2=magnitudeSquared(point-intersection);
					if(dist2<extent2)
					{
						const Real dist=sqrt(dist2);

						if(m_areaSelect)
						{
							Real weight=1.0;
							if(falloff>0.0 && dist>pickRadius)
							{
								weight-=(dist-pickRadius)/falloff;
							}
							if(fuzzy)
							{
								weight-=randomReal(0.0,fuzzy);
							}

							m_colorRim[pointIndex]=m_areaSelect?
									Color(0.0,0.0,weight,1.0): blue;
						}

						//* hidden ignored unless ctrl and shift both held
						U32 item=m_event.item();
						if((closest<0 || closestDist>dist) &&
							((item&WindowEvent::e_keyShift &&
							item&WindowEvent::e_keyControl) ||
							!hidden(pointIndex)))
						{
							closest=pointIndex;
							closestDist=dist;
						}
					}
				}
			}
			if(!m_areaSelect && closest>=0)
			{
				m_colorRim[closest]=blue;
			}

#if FALSE
			spDrawBrush->pushDrawMode(m_spSolid);
			spDrawBrush->draw(m_spDriver,&white);
			spDrawBrush->popDrawMode();
#endif

			if(m_event.isKeyPress(WindowEvent::e_itemAny))
			{
				U32 item=m_event.item();
				BWORD shifted=FALSE;
				if(item&WindowEvent::e_keyShift)
				{
					shifted=TRUE;
					item^=WindowEvent::e_keyShift;
				}
				BWORD ctrled=FALSE;
				if(item&WindowEvent::e_keyControl)
				{
					ctrled=TRUE;
					item^=WindowEvent::e_keyControl;
				}

				if(item==WindowEvent::e_keyCarriageReturn)
				{
					m_areaSelect=!m_areaSelect;
				}

				if(modifyMode=="Drag")
				{
					if(item==WindowEvent::e_keyEscape)
					{
						m_pickArray.clear();
					}

					if(item==WindowEvent::e_keyDelete)
					{
						const I32 pickCount=m_pickArray.size();
						for(I32 pickIndex=0;pickIndex<pickCount;pickIndex++)
						{
							const Pick& rPick=m_pickArray[pickIndex];
							const I32 pointIndex=rPick.m_elementId;

							if(shifted)
							{
								hideToggle(pointIndex);
							}
							else
							{
								revert(pointIndex);
							}
						}
						m_pickArray.clear();
					}
				}
			}

			if(modifyMode=="Drag" && !m_dragging &&
					m_event.isMouseRelease(WindowEvent::e_itemLeft))
			{
				U32 item=m_event.item();
				BWORD shifted=FALSE;
				if(item&WindowEvent::e_keyShift)
				{
					shifted=TRUE;
					item^=WindowEvent::e_keyShift;
				}
				BWORD ctrled=FALSE;
				if(item&WindowEvent::e_keyControl)
				{
					ctrled=TRUE;
					item^=WindowEvent::e_keyControl;
				}

				if(closest>=0)
				{
					if(ctrled)
					{
						//* NOTE can ctrl-shift-pick hidden
						if(shifted || !hidden(closest))
						{
							pickToggle(closest);
						}
					}
					else if(picked(closest)<1.0)
					{
						if(!shifted)
						{
							m_pickArray.clear();
						}
						if(m_areaSelect)
						{
							for(I32 pointIndex=0;pointIndex<outputCount;
									pointIndex++)
							{
								//* HACK using color for region check
								if(m_colorRim[pointIndex][2]>0.0 &&
										!hidden(pointIndex))
								{
									pick(pointIndex,m_colorRim[pointIndex][2]);
								}
							}
						}
						if(!hidden(closest))
						{
							pick(closest,1.0);
						}
					}
				}
				else if(!ctrled)
				{
					m_pickArray.clear();
				}
#if FE_FCO_EDIT_DEBUG
				feLog("FollicleOp::handle pickArray");
				const I32 pickCount=m_pickArray.size();
				for(I32 pickIndex=0;pickIndex<pickCount;pickIndex++)
				{
					const Pick& rPick=m_pickArray[pickIndex];
					const I32 pointIndex=rPick.m_elementId;
					const Real pointWeight=rPick.m_weight;
					feLog(" %d:%.2f",pointIndex,pointWeight);
				}
				feLog("\n");
#endif
			}

			const WindowEvent::MouseButtons buttons=m_event.mouseButtons();
			if(buttons)
			{
				//* TODO screen aware
				const Real moveScale=3e-3;

				//* TODO use last UV to current UV
				const Real uvScale=0.3;

				const I32 mouseX=m_event.mouseX();
				const I32 mouseY=m_event.mouseY();

				const I32 moveX=mouseX-m_lastX;
				const I32 moveY=mouseY-m_lastY;

				const Vector2 uv=spImpact.isValid()?
						spImpact->uv(): Vector2(0.0,0.0);

#if FALSE
				const SpatialVector deltaP=
						moveScale*(deltaX*moveX+deltaY*moveY);

				const SpatialVector du=spImpact->du();
				const SpatialVector dv=spImpact->dv();
				const SpatialVector delta=catalog<bool>("useUV")?
						uvScale*SpatialVector(dot(deltaP,du),dot(deltaP,dv)):
						deltaP;
#else
				const SpatialVector delta=catalog<bool>("useUV")?
						SpatialVector(uv-m_lastUV):
						moveScale*(deltaX*moveX+deltaY*moveY);
#endif

				if(modifyMode=="Drag")
				{
					if(buttons&WindowEvent::e_mbLeft)
					{
						if(m_event.isMouseDrag())
						{
							if(closest>=0)
							{
								movePicked(delta);
							}
						}
					}

					if(buttons&WindowEvent::e_mbMiddle)
					{
						if(impacted && m_event.isMousePress())
						{
							addFollicle(intersection);
						}
					}
				}
				else if(modifyMode=="Paint")
				{
					if(closest>=0)
					{
						if(action=="Apply")
						{
							if(buttons&WindowEvent::e_mbLeft)
							{
								paintValue(intensity);
							}
							else if(buttons&WindowEvent::e_mbMiddle)
							{
								paintValue(-intensity);
							}
						}
						else if(action=="Smooth")
						{
							if(buttons&WindowEvent::e_mbLeft)
							{
								smoothValue(intensity);
							}
						}
						else if(action=="Restore")
						{
							if(buttons&WindowEvent::e_mbLeft)
							{
								restoreValue(intensity);
							}
						}
					}
				}

				m_lastX=mouseX;
				m_lastY=mouseY;
				m_lastUV=uv;
				m_dragging=m_event.isMouseDrag();
			}

			//* TODO param
			const Real mapX=10.0;
			const Real mapY=200.0;
			const Real mapScale=400.0;

			const Real circleRadius=3.0;
			const SpatialVector circleScale(
					circleRadius,circleRadius,circleRadius);

			const Real sampleRadius=4.0;
			const SpatialVector sampleScale(
					sampleRadius,sampleRadius,sampleRadius);

			SpatialTransform circleTransform;
			setIdentity(circleTransform);

			if(impacted)
			{
				const Real offset=0.01;	//TODO param
				const I32 depth=0;
				const SpatialVector center=intersection+offset*hitNormal;
				drawPicker(spDrawBrush,center,up,hitNormal,
						m_areaSelect? pickRadius: singleRadius,
						depth,&yellow);
				if(m_areaSelect)
				{
					drawPicker(spDrawBrush,center,up,hitNormal,
							falloff>0.0? pickRadius+falloff: 1.05*pickRadius,
							depth,
							falloff>0.0? &orange: &black);
				}

				const Vector2 uv=spImpact->uv();
				const SpatialVector pixel(mapX+mapScale*uv[0],
						mapY+mapScale*uv[1]);

				setTranslation(circleTransform,pixel);
				spDrawOverlay->drawCircle(circleTransform,&circleScale,magenta);
			}

			for(U32 pass=0;pass<2;pass++)
			{
				sp<RecordArray> spRA=
						m_spState->getArray(pass? "Move": "Hide");
				if(spRA.isValid())
				{
					for(int i=0;i<spRA->length();i++)
					{
						const I32 pointIndex=m_aIndex(spRA,i);
						if(pointIndex<0 || pointIndex>=outputCount)
						{
							continue;
						}

						if(!pass && moved(pointIndex))
						{
							//* will draw with moved points
							continue;
						}

						const SpatialVector point=pass?
								m_spInputPoint->spatialVector(pointIndex):
								m_spOutputPoint->spatialVector(pointIndex);
						sp<SurfaceI::ImpactI> spImpact=
								m_spDriver->nearestPoint(point);
						if(!spImpact.isValid())
						{
							continue;
						}

						const Vector2 origUV=spImpact->uv();
						const Vector2 uv=pass? origUV+m_aUV(spRA,i): origUV;
						const SpatialVector pixel(mapX+mapScale*uv[0],
								mapY+mapScale*uv[1]);

#if FALSE
						String text;
//						text.sPrintf("%.2f %.2f",uv[0],uv[1]);
						text.sPrintf("%d",pointIndex);
						spDrawOverlay->drawAlignedText(pixel,text,blue);
#endif

						const BWORD move=(pass && !hidden(pointIndex));

						setTranslation(circleTransform,pixel);
						if(move)
						{
							const Real value=m_aValue(spRA,i);
							Color color(1.0,value,0.0);

							if(modifyMode=="Paint")
							{
								m_colorCenter[pointIndex]=color;
							}
						}
						else
						{
							spDrawOverlay->drawCircle(circleTransform,
									&circleScale,move? cyan: red);
						}
					}
				}
			}

			sp<RecordArray> spRA=m_spState->getArray("Add");
			if(spRA.isValid())
			{
				for(int i=0;i<spRA->length();i++)
				{
					const Vector2 uv=m_aUV(spRA,i);
					const SpatialVector pixel(mapX+mapScale*uv[0],
							mapY+mapScale*uv[1]);

					const Real value=m_aValue(spRA,i);
					Color color(1.0,value,0.0);

					setTranslation(circleTransform,pixel);
					spDrawOverlay->drawCircle(circleTransform,
							&circleScale,sparkle);
				}
			}

			if(m_spOutputAttribute.isValid())
			{
				spDrawBrush->pushDrawMode(m_spSolid);

				std::vector<SpatialVector> locationArray(outputCount);
				std::vector<SpatialVector> normalArray(outputCount);
				std::vector<Vector2> uvArray(outputCount);
				std::vector<Vector2> pixelArray(outputCount);
				std::vector<Color> colorArray(outputCount);

				for(I32 pointIndex=0;pointIndex<outputCount;pointIndex++)
				{
					const SpatialVector point=
							m_spOutputPoint->spatialVector(pointIndex);

					//* TODO read UV attribute (also need to update those uvs)
					sp<SurfaceI::ImpactI> spImpact=
							m_spDriver->nearestPoint(point);
					if(!spImpact.isValid())
					{
						continue;
					}

					const Vector2 uv=spImpact->uv();
					const SpatialVector pixel(mapX+mapScale*uv[0],
							mapY+mapScale*uv[1]);
					setTranslation(circleTransform,pixel);

					const Real value=
							m_spOutputAttribute->real(pointIndex);
					const Color sampleColor(value,value,value);
#if FALSE
					spDrawOverlay->drawCircle(circleTransform,
							&sampleScale,sampleColor);
#else
					spDrawOverlay->drawCircle(circleTransform,
							&circleScale,blue);
//					spDrawOverlay->drawPoints(&pixel,NULL,1,
//							FALSE,&blue);
#endif

					locationArray[pointIndex]=spImpact->intersection();
					normalArray[pointIndex]=spImpact->normal();
					uvArray[pointIndex]=uv;
					pixelArray[pointIndex]=pixel;
					colorArray[pointIndex]=sampleColor;
				}

				//* Delaunay
				std::vector<Vector3i> triangleArray;
				PointConnect::solve(uvArray,triangleArray);

				const U32 primitiveCount=triangleArray.size();
				for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;
						primitiveIndex++)
				{
					const Vector3i& rTriangle=triangleArray[primitiveIndex];
					const I32 tri0=rTriangle[0];
					const I32 tri1=rTriangle[1];
					const I32 tri2=rTriangle[2];

					SpatialVector vertex[3];
					vertex[0]=locationArray[tri0];
					vertex[1]=locationArray[tri1];
					vertex[2]=locationArray[tri2];

					SpatialVector vertexNormal[3];
					vertexNormal[0]=normalArray[tri0];
					vertexNormal[1]=normalArray[tri1];
					vertexNormal[2]=normalArray[tri2];

					Color vertexColor[3];
					vertexColor[0]=colorArray[tri0];
					vertexColor[1]=colorArray[tri1];
					vertexColor[2]=colorArray[tri2];

					spDrawBrush->drawTriangles(vertex,vertexNormal,
							NULL,3,DrawI::e_discrete,
							TRUE,vertexColor);

					vertex[0]=pixelArray[tri0];
					vertex[1]=pixelArray[tri1];
					vertex[2]=pixelArray[tri2];

					set(vertexNormal[0],0.0,0.0,1.0);
					set(vertexNormal[1],0.0,0.0,1.0);
					set(vertexNormal[2],0.0,0.0,1.0);

					spDrawOverlay->drawTriangles(vertex,vertexNormal,
							NULL,3,DrawI::e_discrete,
							TRUE,vertexColor);
				}

				spDrawBrush->popDrawMode();
			}
		}

		const I32 outputCount=m_spOutputPoint->count();
		if(modifyMode=="Drag")
		{
			const I32 pickCount=m_pickArray.size();
			for(I32 pickIndex=0;pickIndex<pickCount;pickIndex++)
			{
				const Pick& rPick=m_pickArray[pickIndex];
				const I32 pointIndex=rPick.m_elementId;
				const Real pointWeight=rPick.m_weight;
				if(pointIndex>=0 && pointIndex<outputCount)
				{
					m_colorCenter[pointIndex]=Color(1.0,pointWeight,0.0,1.0);
				}
			}
		}

		sp<RecordArray> spRA=m_spState->getArray("Hide");
		if(spRA.isValid())
		{
			for(int i=0;i<spRA->length();i++)
			{
				const I32 pointIndex=m_aIndex(spRA,i);
				if(pointIndex>=0 && pointIndex<outputCount)
				{
					m_colorRim[pointIndex]=red;
				}
			}
		}

		for(I32 pointIndex=0;pointIndex<outputCount;pointIndex++)
		{
			const Color& rColorCenter=m_colorCenter[pointIndex];
			const Color& rColorRim=m_colorRim[pointIndex];
			if(rColorCenter[3]<=0.0 && rColorRim[3]<=0.0)
			{
				continue;
			}

			const SpatialVector point=
					m_spOutputPoint->spatialVector(pointIndex);

			//* TODO cached 'moved', which is a search

			const Color shadow=cyan;
			drawSquare(spDrawOverlay,point,rimSize,
					rColorCenter,rColorRim,
					moved(pointIndex)? &shadow: NULL, NULL);
		}

		const SpatialVector textPoint(8,48);
		String text;
		text.sPrintf("%s Select",m_areaSelect? "Area": "Single");
		spDrawOverlay->drawAlignedText(textPoint,text,blue);

		spRA=m_spState->getArray("Add");
		if(spRA.isValid())
		{
#if FE_FCO_OUTPUT_DEBUG
			feLog("FollicleOp::handle add count %d\n",spRA->length());
#endif

			for(int i=0;i<spRA->length();i++)
			{
				Record edit=spRA->getRecord(i);

				const SpatialVector point=m_aLocation(edit);

				drawSquare(spDrawOverlay,point,rimSize,
						blank,grey,NULL,&sparkle);
			}
		}

		spDrawOverlay->popDrawMode();
		spDrawBrush->popDrawMode();

		return;
	}

	sp<SurfaceAccessibleI> spInputAccessible;
	if(!access(spInputAccessible,"Input Surface")) return;

	sp<SurfaceAccessibleI> spOutputAccessible;
	if(!accessOutput(spOutputAccessible,a_rSignal)) return;

	if(!access(m_spInputPoint,spInputAccessible,e_point,e_position)) return;

	if(!access(m_spOutputPoint,spOutputAccessible,e_point,e_position)) return;

	if(!access(m_spOutputNormal,spOutputAccessible,e_point,e_normal)) return;

	m_spInputAttribute=NULL;
	m_spOutputAttribute=NULL;
	if(affectAttribute)
	{
		if(!access(m_spInputAttribute,spInputAccessible,e_point,attribute))
		{
			catalog<String>("warning")+=
					"missing point attribute "+attribute+";";
		}
		access(m_spOutputAttribute,spOutputAccessible,
				e_point,attribute,e_quiet);
	}

	const BWORD paramChanged=!m_brushed;
	m_brushed=FALSE;

	if(paramChanged)
	{
		if(!access(m_spDriver,"Driver Surface")) return;
	}

	//* TODO use
	const BWORD rebuild=(paramChanged ||
			catalog<bool>("Output Surface","replaced"));

	const String hideGroup=catalog<String>("hideGroup");
	if(!hideGroup.empty())
	{
		if(!accessOutput(m_spHideGroup,a_rSignal,
				e_pointGroup,hideGroup)) return;

		const U32 outputCount=m_spOutputPoint->count();

		sp<RecordArray> spRA=m_spState->getArray("Hide");
		if(spRA.isValid())
		{
			for(int i=0;i<spRA->length();i++)
			{
				const I32 pointIndex=m_aIndex(spRA,i);
				if(pointIndex>=0 && pointIndex<outputCount)
				{
					m_spHideGroup->append(pointIndex);
				}
			}
		}

	}

	sp<RecordArray> spRA=m_spState->getArray("Move");
	if(spRA.isValid())
	{
#if FE_FCO_OUTPUT_DEBUG
		feLog("FollicleOp::handle move count %d\n",spRA->length());
#endif
		for(int i=0;i<spRA->length();i++)
		{
			const I32 pointIndex=m_aIndex(spRA,i);

			Record edit=spRA->getRecord(i);

#if FE_FCO_OUTPUT_DEBUG
			feLog("FollicleOp::handle move %d/%d prim %d/%d\n",
					i,spRA->length(),pointIndex,m_spOutputPoint->count());
#endif

			const Real original=m_spInputAttribute.isValid()?
					m_spInputAttribute->real(pointIndex): 0.0;

			if(magnitudeSquared(m_aLocation(edit))<1e-6 &&
					fabs(m_aValue(edit)-original)<1e-3)
			{
				//* remove non-contributing record
				spRA->remove(i);
				i--;
				continue;
			}

			const SpatialVector inputPoint=
					m_spInputPoint->spatialVector(pointIndex);

			const SpatialVector outputPoint=inputPoint+m_aLocation(edit);

			m_spOutputPoint->set(pointIndex,outputPoint);

			if(m_spOutputAttribute.isValid())
			{
				m_spOutputAttribute->set(pointIndex,m_aValue(edit));
			}
		}
	}

	sp<DrawI> spOutputDrawI;
	if(!accessDraw(spOutputDrawI,a_rSignal)) return;

	spRA=m_spState->getArray("Add");
	if(spRA.isValid())
	{
#if FE_FCO_OUTPUT_DEBUG
		feLog("FollicleOp::handle append count %d\n",spRA->length());
#endif

		for(int i=0;i<spRA->length();i++)
		{
			Record edit=spRA->getRecord(i);

#if FE_FCO_OUTPUT_DEBUG
			const I32 pointIndex=m_aIndex(spRA,i);
			feLog("FollicleOp::handle append %d/%d point %d/%d\n",
					i,spRA->length(),pointIndex,m_spOutputPoint->count());
#endif

			const SpatialVector point=m_aLocation(edit);

			//* add point to output
			spOutputDrawI->drawPoints(&point,NULL,1,FALSE,NULL);
		}
	}

	return;
}

BWORD FollicleOp::revert(I32 a_index)
{
	BWORD result=FALSE;
	if(hidden(a_index))
	{
		hideToggle(a_index);
		result=TRUE;
	}
	sp<RecordArray> spRA=m_spState->getArray("Move");
	if(spRA.isValid())
	{
		for(int i=0;i<spRA->length();i++)
		{
			const I32 pointIndex=m_aIndex(spRA,i);
			if(a_index==pointIndex)
			{
				spRA->remove(i);
				return TRUE;
			}
		}
	}
	return result;
}

void FollicleOp::setValueLimited(Record a_edit,Real a_value)
{
	//* TODO param min,max
	if(a_value>1.0)
	{
		a_value=1.0;
	}
	if(a_value<0.0)
	{
		a_value=0.0;
	}
	m_aValue(a_edit)=a_value;
}

void FollicleOp::paintValue(Real a_intensity)
{
	const U32 outputCount=m_spOutputPoint->count();
	for(I32 pointIndex=0;pointIndex<outputCount;pointIndex++)
	{
		const Real weight=m_colorRim[pointIndex][2];
		if(weight<=0.0)
		{
			continue;
		}

		Record edit=getEdit(pointIndex);
		if(!m_aValue.check(edit))
		{
			continue;
		}

		setValueLimited(edit,m_aValue(edit)+a_intensity*weight);
	}
}

void FollicleOp::smoothValue(Real a_intensity)
{
	const U32 outputCount=m_spOutputPoint->count();

	Real sum=0.0;
	Real totalWeight=0.0;
	for(I32 pointIndex=0;pointIndex<outputCount;pointIndex++)
	{
		const Real weight=m_colorRim[pointIndex][2];
		if(weight<=0.0)
		{
			continue;
		}

		Record edit=getEdit(pointIndex);
		if(!m_aValue.check(edit))
		{
			continue;
		}

		sum+=m_aValue(edit)*weight;
		totalWeight+=weight;

//		feLog("FollicleOp::smoothValue value %.6G weight %.6G\n",
//				m_aValue(edit),weight);
	}
	if(totalWeight<=1e-6)
	{
		return;
	}

	const Real target=sum/totalWeight;

//	feLog("FollicleOp::smoothValue sum %.6G total %.6G target %.6G\n",
//			sum,totalWeight,target);

//	sum=0.0;

	for(I32 pointIndex=0;pointIndex<outputCount;pointIndex++)
	{
		const Real weight=m_colorRim[pointIndex][2];
		if(weight<=0.0)
		{
			continue;
		}

		Record edit=getEdit(pointIndex);
		if(!m_aValue.check(edit))
		{
			continue;
		}

		const Real ratio=a_intensity*weight;

		setValueLimited(edit,(1.0-ratio)*m_aValue(edit)+ratio*target);

//		sum+=m_aValue(edit)*weight;
	}

//	feLog("FollicleOp::smoothValue new sum %.6G mean %.6G\n",
//			sum,sum/totalWeight);
}

void FollicleOp::restoreValue(Real a_intensity)
{
	const U32 outputCount=m_spOutputPoint->count();
	for(I32 pointIndex=0;pointIndex<outputCount;pointIndex++)
	{
		const Real weight=m_colorRim[pointIndex][2];
		if(weight<=0.0)
		{
			continue;
		}

		Record edit=getEdit(pointIndex);
		if(!m_aValue.check(edit))
		{
			continue;
		}

		const Real ratio=a_intensity*weight;
		const Real original=m_spInputAttribute.isValid()?
				m_spInputAttribute->real(pointIndex): 0.0;

		setValueLimited(edit,(1.0-ratio)*m_aValue(edit)+ratio*original);
	}
}

BWORD FollicleOp::hideToggle(I32 a_index)
{
	sp<RecordArray> spRA=m_spState->getArray("Hide");
	if(spRA.isValid())
	{
		for(int i=0;i<spRA->length();i++)
		{
			const I32 pointIndex=m_aIndex(spRA,i);
			if(a_index==pointIndex)
			{
				spRA->remove(i);
				return TRUE;
			}
		}
	}

	hide(a_index);
	return FALSE;
}

void FollicleOp::hide(I32 a_index)
{
	//* remove if point was an add
	sp<RecordArray> spRA=m_spState->getArray("Add");
	if(spRA.isValid())
	{
		for(int i=0;i<spRA->length();i++)
		{
			const I32 pointIndex=m_aIndex(spRA,i);
			if(a_index==pointIndex)
			{
				spRA->remove(i);
//				Record record=spRA->getRecord(i);
//				m_spState->remove(record);
				return;
			}
		}
	}
	//* no change if already hidden
	spRA=m_spState->getArray("Hide");
	if(spRA.isValid())
	{
		for(int i=0;i<spRA->length();i++)
		{
			const I32 pointIndex=m_aIndex(spRA,i);
			if(a_index==pointIndex)
			{
				return;
			}
		}
	}
	Record hide=m_spScope->createRecord("Hide");
	m_aIndex(hide)=a_index;

	m_spState->add(hide);
}

BWORD FollicleOp::hidden(I32 a_index)
{
	sp<RecordArray> spRA=m_spState->getArray("Hide");
	if(spRA.isValid())
	{
		for(int i=0;i<spRA->length();i++)
		{
			const I32 pointIndex=m_aIndex(spRA,i);
			if(a_index==pointIndex)
			{
				return TRUE;
			}
		}
	}
	return FALSE;
}

BWORD FollicleOp::moved(I32 a_index)
{
	sp<RecordArray> spRA=m_spState->getArray("Move");
	if(spRA.isValid())
	{
		for(int i=0;i<spRA->length();i++)
		{
			const I32 pointIndex=m_aIndex(spRA,i);
			if(a_index==pointIndex)
			{
				return TRUE;
			}
		}
	}
	return FALSE;
}

Record FollicleOp::getEdit(I32 a_index)
{
	Record edit=m_spEditMap->lookup(a_index);
	if(!edit.isValid())
	{
		edit=m_spScope->createRecord("Move");
		m_aIndex(edit)=a_index;
		m_aValue(edit)=m_spInputAttribute.isValid()?
				m_spInputAttribute->real(a_index): 0.0;

		m_spState->add(edit);
	}

	return edit;
}

void FollicleOp::addFollicle(const SpatialVector& a_intersection)
{
	U32 maxIndex=m_spOutputPoint->count();

	const I32 pickCount=m_pickArray.size();
	for(I32 pickIndex=0;pickIndex<pickCount;pickIndex++)
	{
		const Pick& rPick=m_pickArray[pickIndex];
		const I32 pointIndex=rPick.m_elementId;
		if(maxIndex<pointIndex)
		{
			maxIndex=pointIndex;
		}
	}

	Record edit=m_spScope->createRecord("Add");
	m_aIndex(edit)=maxIndex;
	m_aLocation(edit)=a_intersection;
	m_aValue(edit)=0.0;

	sp<SurfaceI::ImpactI> spImpact=m_spDriver->nearestPoint(a_intersection);
	m_aUV(edit)=spImpact.isValid()? spImpact->uv(): Vector2(0.0,0.0);

	m_spState->add(edit);
}

void FollicleOp::movePicked(const SpatialVector& a_rDelta)
{
#if FE_FCO_EDIT_DEBUG
	feLog("FollicleOp::movePicked %s\n",c_print(a_rDelta));
#endif

	const BWORD useUV=catalog<bool>("useUV");

	const I32 pickCount=m_pickArray.size();
	for(I32 pickIndex=0;pickIndex<pickCount;pickIndex++)
	{
		const Pick& rPick=m_pickArray[pickIndex];
		const I32 pointIndex=rPick.m_elementId;
		const Real pointWeight=rPick.m_weight;
		Record edit=getEdit(pointIndex);
		if(useUV)
		{
			slideOnUV(pointIndex,a_rDelta*pointWeight);
		}
		else
		{
			m_aLocation(edit)+=a_rDelta*pointWeight;

#if FE_FCO_EDIT_DEBUG
			SpatialVector inputPoint;
			set(inputPoint);
			const U32 inputCount=m_spInputPoint->count();
			if(pointIndex<inputCount)
			{
				inputPoint=m_spInputPoint->spatialVector(pointIndex);
			}

			const SpatialVector outputPoint=inputPoint+m_aLocation(edit);

			feLog("FollicleOp::movePicked %d %d at %s by %s to %s\n",
					pickIndex,pointIndex,c_print(inputPoint),
					c_print(m_aLocation(edit)),c_print(outputPoint));
#endif

			snapToSurface(pointIndex);

#if FE_FCO_EDIT_DEBUG
			feLog("  snap %s\n",c_print(m_aLocation(edit)));
#endif
		}
	}
}

void FollicleOp::slideOnUV(I32 a_index,const Vector2 a_deltaUV)
{
#if FE_FCO_EDIT_DEBUG
	feLog("FollicleOp::slideOnUV %d %s\n",a_index,c_print(a_deltaUV));
#endif

	const U32 inputCount=m_spInputPoint->count();

	SpatialVector point;
	set(point);

	Vector2 origUV;
	set(origUV);

	if(a_index<inputCount)
	{
		point=m_spInputPoint->spatialVector(a_index);

		sp<SurfaceI::ImpactI> spImpact=m_spDriver->nearestPoint(point);
		if(!spImpact.isValid())
		{
			return;
		}
		origUV=spImpact->uv();
	}

	Record edit=getEdit(a_index);

	m_aUV(edit)+=a_deltaUV;
	Vector2 uv=origUV+m_aUV(edit);

	const SpatialVector sample=m_spDriver->samplePoint(uv);

#if FE_FCO_EDIT_DEBUG
	feLog("  orig %s add %s uv %s\n",
			c_print(origUV),c_print(m_aUV(edit)),c_print(uv));
	feLog("  point %s sample %s\n",
			c_print(point),c_print(sample));
	feLog("  was %s now %s\n",
			c_print(m_aLocation(edit)),c_print(sample-point));
#endif

	m_aLocation(edit)=sample-point;
}

void FollicleOp::snapToSurface(I32 a_index)
{
	const U32 inputCount=m_spInputPoint->count();

	SpatialVector point;
	set(point);
	if(a_index<inputCount)
	{
		point=m_spInputPoint->spatialVector(a_index);
	}

	Record edit=getEdit(a_index);

	sp<SurfaceI::ImpactI> spImpact=
			m_spDriver->nearestPoint(point+m_aLocation(edit));
	if(!spImpact.isValid())
	{
		return;
	}
	const SpatialVector intersection=spImpact->intersection();

	m_aLocation(edit)=intersection-point;
}

void FollicleOp::setupState(void)
{
	m_aIndex.populate(m_spScope,"Move","Index");
	m_aLocation.populate(m_spScope,"Move","Location");
	m_aUV.populate(m_spScope,"Move","UV");
	m_aValue.populate(m_spScope,"Move","Value");

	m_aIndex.populate(m_spScope,"Add","Index");
	m_aLocation.populate(m_spScope,"Add","Location");
	m_aUV.populate(m_spScope,"Add","UV");
	m_aValue.populate(m_spScope,"Add","Value");

	m_aIndex.populate(m_spScope,"Hide","Index");

	m_spEditMap=m_spState->lookupMap(m_aIndex);
}

BWORD FollicleOp::loadState(const String& rBuffer)
{
	const BWORD success=OperatorState::loadState(rBuffer);

	m_spEditMap=m_spState->lookupMap(m_aIndex);

	return success;
}
